// fournit le premier milliard de chiffres décimaux de pi en 30 minutes environ sur
// macBook Pro M1 2020 16 Go (macOS Ventura 13.5, Xcode 14.3.1 (14E300c), gmp-6.2.99)
// macBook Pro M1 2020 16 Go (macOS Sonoma 14.1, Xcode 15.0.1 (15A507), gmp-6.3.0)
// ligne de commande (avec installation par défaut de gmp) :
// g++ -std=c++17 pi.cpp -opi++ -lgmp -lgmpxx
// ou (avec installation statique de gmp dans <prefix>)
// g++ -std=c++17 pi.cpp -opi++ -I<prefix>/include -L<prefix>/lib -lgmp -lgmpxx
// utilisable aussi sur plateforme GNU/Linux comme ci-dessus
// ou sur plateforme MSYS2/MinGW (Windows) : comme ci-dessus en ajoutant -D_MINGW_ à la ligne de commande

#include <stdio.h> // FILE, fopen(), fclose() ...
#include <gmpxx.h> // gmp c++
#include <iostream> // std::cout << ... << std::endl
#include <filesystem> // std::filesystem::current_path() ( -std=c++17 )
#include <iomanip> // std::fixed, std::setprecision()
#include <fstream>
#include <sstream> // std::stringstream
#include <thread> // std::thread
#include <chrono> // std::chrono::system_clock()
#ifdef _MINGW_
#include <windows.h>
#endif

using namespace std;

static unsigned long N; // nombre de décimales de pi calculées (argv[1] de main())
static string pi_path; // chemin du répertoire "pi"
static void pi(); // fonction définie en fin de fichier appelée par main()

int main (int argc, char* argv[])
{
#ifdef _MINGW_
    SetConsoleOutputCP(65001);
    stringstream ss;
    ss << std::filesystem::current_path();
    pi_path = ss.str().substr(1, ss.str().length() - 2) + "\\\\pi";
#else
    pi_path = string(filesystem::current_path())+"/pi";
#endif
    filesystem::directory_entry entry(pi_path);
    if (!entry.exists()){
        if (!filesystem::create_directory("pi")) {
            cout << "la création du répertoire " << pi_path << " a échoué" << endl;
            exit(EXIT_FAILURE);
        }
    }
    if (argc != 2) {
        std:: cout << "synopsis: " << argv[0] << " < nombre de décimales de pi >\n";
        return 0;
    }
    N = std::stol(argv[1]);
    pi();
}

//---- entrées-sorties

void bigint_to_binfile(const mpz_class bigint, const string& fName)
{
    FILE* f = fopen(fName.c_str(),"w");
#ifdef _MINGW_
    mpz_out_str(f,2,bigint.get_mpz_t());
#else
    mpz_out_raw(f,bigint.get_mpz_t());
#endif
    fclose(f);
}

void bigint_from_binfile(mpz_class &bigint, const string& fName)
{
    FILE* f = fopen(fName.c_str(),"r");
#ifdef _MINGW_
    mpz_inp_str(bigint.get_mpz_t(),f,2);
#else
    mpz_inp_raw (bigint.get_mpz_t(),f);
#endif
    fclose(f);
}

//---- temps courant en secondes

namespace std::chrono {
    double secondes() {
        return duration_cast< microseconds >(system_clock::now().time_since_epoch()).count()/1000000.;
    }
}

double secondes()
{
    return std::chrono::secondes();
}

//---- classe représentant la fonction affine x -> u/w + x * v/w avec u,v,w grands entiers de type mpz_class

class affine {
public:
    mpz_class u,v,w;
    affine(){}
    affine(const string& fName){
        bigint_from_binfile(u,fName+"_a");
        bigint_from_binfile(v,fName+"_b");
        bigint_from_binfile(w,fName+"_c");
    }
    void compose(const affine& af){
        u *=af.w;
        u += v*af.u;
        v *= af.v;
        w *= af.w;
    }
    void reduis(){
        reduis(u.get_mpz_t(),v.get_mpz_t(),w.get_mpz_t());
    }
    void to_binfiles(const string& fName){
        bigint_to_binfile(u,fName+"_a");
        bigint_to_binfile(v,fName+"_b");
        bigint_to_binfile(w,fName+"_c");
    }
private:
    void reduis(mpz_t u, mpz_t v, mpz_t w) // division de u,v,w par leur PGCD
    {
        mpz_t t;
        mpz_init(t);
        mpz_gcd(t,u,v);
        mpz_gcd(t,t,w);
        mpz_divexact(u,u,t);
        mpz_divexact(v,v,t);
        mpz_divexact(w,w,t);
        mpz_clear(t);
    }
};

//---- composée de plusieurs fonctions affines, avec multithreading (paramètre i > 0)

void compose(affine& af, affine (&f)(unsigned long), unsigned long a, unsigned long b, int i = 0)
// af = f(a)of(a+1)o...of(b-1)  (a < b requis)
{
    unsigned long mil = (a+b)/2;
    if (b == a+1) {
        af = f(a);
    }
    else {
        affine af2;
        if (i > 1) {
            thread th(compose,ref(af),ref(f),a,mil,i-1);
            compose(af2,f,mil,b,i-1);
            th.join();
        }
        else if (i == 1) {
            compose(af,f,a,mil);
            compose(af2,f,mil,b);
            thread th([&]{af.reduis();});
            af2.reduis();
            th.join();
        }
        else {
            compose(af,f,a,mil);
            compose(af2,f,mil,b);
        }
        af.compose(af2);
    }
}

void etape(int j, affine (&f)(unsigned long), unsigned long a, unsigned long b) // (a < b requis)
{
    cout << "étape " << j << endl;
    
    double sec1 = secondes();
    affine af;
    compose(af,f,a,b,4);

    stringstream s;
    s << "pi/etape" << j << "_" << N;
    af.to_binfiles(s.str());

    cout << "durée: " << fixed << setprecision(3) << secondes()-sec1 << " s" << endl;
}

//---- cf. le document pdf joint pour les notations suivantes a(p), b(p), f(p), g(p)

// calcul par tranches de m termes (m impair) de la somme de la série de terme général a(n)c(n)
// ( b(n)=c(n+1)/c(n) )

/*
static const int m = 1;
mpz_class poly[4*m] = {1640_mpz, 3066_mpz, 1804_mpz, 330_mpz};
*/

static const int m = 63;
mpz_class poly[4*m] =
#include "poly63.txt" // liste des coefficients du polynôme de degré 4m-1 'poly'

mpz_class horner(mpz_class poly[], int d, mpz_class p)
// Calcul de poly[0]*p^(d-1)+poly[1]*p^(d-2)+...+poly[d-1] par la méthode de Horner
{
    mpz_class x = poly[0];
    for (int i=1; i<d; i++){
        x *=p;
        x +=poly[i];
    }
    return x;
}

mpz_class a(mpz_class p)
{
    return horner(poly,4*m,p);
}

mpq_class b(unsigned long p)
{
    mpz_class num = 1;
    mpz_class den = 1;
    for (unsigned long j = 4*m*p+1; j<=4*m*(p+1); j++) {
        num *= j;
        den *= 2*j+8*m-1;
    }
    den <<= 6*m;
    den = -den;
    mpq_class frac(num,den);
    frac.canonicalize();
    return frac;
}

affine f(unsigned long p) // f : x -> a(p)+b(p)*x
{
    affine af;
    af.u = a(p);
    mpq_class frac = b(p);
    af.v = frac.get_num();
    af.w = frac.get_den();
    af.u *= af.w;
    return af;
}

mpz_class dixExpN; // valeur 10^N (initialisée en parallèle ci-dessous dans g())

affine g(unsigned long p) // g(p) = f(0)of(1)o...of(p) . (avec p = max(N/(3*m),2))
{
    cout << "composée affine en deux étapes enregistrées:" << endl;
    etape(1,f,0,(p/2)); // (0 < p/2)
    etape(2,f,(p/2),(p+1)); // (p/2 < p+1)

    cout << "composition des deux étapes..." << endl;
    double sec1=secondes();

    stringstream s1;
    s1 << "pi/etape1_" << N;
    affine af(s1.str());

    stringstream s2;
    s2 << "pi/etape2_" << N;
    affine af2(s2.str());
    
    thread th([](){mpz_ui_pow_ui(dixExpN.get_mpz_t(),10,N);}); //initialisation de dixExpN
    af.compose(af2);
    th.join();

    cout << "durée: " << fixed << setprecision(3) << secondes()-sec1 << " s" << endl;
    return af;
}

mpz_class c0(unsigned int m) // c_m(0) dans le document pdf joint
{
    mpz_class r = 1;
    mpz_2fac_ui(r.get_mpz_t(),8*m-1);
    r <<= 3*(m-1);
    return r;
}

void pi()
{
    double sec1 = secondes();
    unsigned long p = N/(3*m);
    if (p < 2) p = 2;
    affine af = g(p);
    
    double sec = secondes();
    cout << "\nprécision..." << endl;
    mpf_class x(mpq_class(a(p+1)*af.v,c0(m)*af.w));
    cout << "nombre de termes de la somme partielle = " << p <<endl;
    cout << "incertitude = " << scientific << x << endl;
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec << endl;

    cout << "valeur approchée de pi par défaut avec " << N << " décimales..." << endl;
    double sec2 = secondes();
    mpz_class res = (dixExpN*af.u)/(c0(m)*af.w);
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec2 << endl;
    cout << N << " décimales de pi en " << fixed << setprecision(3) << secondes()-sec1 << " s" << endl;

    cout << "\nenregistrement binaire..." << endl;
    double sec3 = secondes();
    stringstream titre;
    titre << "pi-" << N;
    bigint_to_binfile(res,string("pi/")+titre.str());
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec3 << " s" << endl;
    cout << "durée totale: " << fixed << setprecision(3) << secondes()-sec1 << " s" << endl;
    
    cout << "\nenregistrement décimal..." << endl;
    mpz_class res1; //pour vérification du nombre N de décimales ( res == res1)
    thread th([&] {res1 = (dixExpN*(af.u+a(p+1)*af.v))/(c0(m)*af.w);});
    double sec4 = secondes();
    string strres = res.get_str();
    if (N <= 1000000) cout << strres.substr(0,1) << "." << strres.substr(1) << endl;
    ofstream fs;
    fs.open(pi_path+"/"+titre.str()+".txt");
    fs << strres.substr(0,1) << "." << strres.substr(1);
    fs.close();
    cout << "durée: " << fixed << setprecision(3) << secondes()-sec4 << " s" << endl;
    cout << "durée totale: " << fixed << setprecision(3) << secondes()-sec1 << " s" << endl;
    th.join();
    if (res == res1)
    cout << "résultat décimal dans le fichier " << titre.str() << ".txt du répertoire " << pi_path << endl;
    else cout << "résultat décimal INEXACT dans le fichier " << titre.str() << ".txt du répertoire " << pi_path << endl;
}
